import allin1

from pathlib import Path

CWD = Path(__file__).resolve().parent


def test_sonify():
  result = allin1.AnalysisResult.from_json(CWD / 'test.json')
  allin1.sonify(result)


def test_sonify_save():
  result = allin1.AnalysisResult.from_json(CWD / 'test.json')
  allin1.sonify(
    result,
    out_dir='./sonif'
  )



import numpy as np
import json
import torch

from dataclasses import asdict
from pathlib import Path
from typing import List, Union
from .utils import mkpath, compact_json_number_array
from .typings import AllInOneOutput, AnalysisResult, PathLike


def compute_activations(logits: AllInOneOutput):
  activations_beat = torch.sigmoid(logits.logits_beat[0]).cpu().numpy()
  activations_downbeat = torch.sigmoid(logits.logits_downbeat[0]).cpu().numpy()
  activations_segment = torch.sigmoid(logits.logits_section[0]).cpu().numpy()
  activations_label = torch.softmax(logits.logits_function[0], dim=0).cpu().numpy()
  return {
    'beat': activations_beat,
    'downbeat': activations_downbeat,
    'segment': activations_segment,
    'label': activations_label,
  }


def expand_paths(paths: List[Path]):
  expanded_paths = set()
  for path in paths:
    if '*' in str(path) or '?' in str(path):
      matches = list(path.parent.glob(path.name))
      if not matches:
        raise FileNotFoundError(f'Could not find any files matching {path}')
      expanded_paths.update(matches)
    else:
      expanded_paths.add(path)

  return list(expanded_paths)


def check_paths(paths: List[Path]):
  missing_files = []
  for path in paths:
    if not path.is_file():
      missing_files.append(str(path))
  if missing_files:
    raise FileNotFoundError(f'Could not find the following files: {missing_files}')


def rmdir_if_empty(path: Path):
  try:
    path.rmdir()
  except (FileNotFoundError, OSError):
    pass


def save_results(
  results: Union[AnalysisResult, List[AnalysisResult]],
  out_dir: PathLike,
):
  if not isinstance(results, list):
    results = [results]

  out_dir = mkpath(out_dir)
  out_dir.mkdir(parents=True, exist_ok=True)
  for result in results:
    out_path = out_dir / result.path.with_suffix('.json').name
    result = asdict(result)
    result['path'] = str(result['path'])

    activations = result.pop('activations')
    if activations is not None:
      np.savez(str(out_path.with_suffix('.activ.npz')), **activations)

    embeddings = result.pop('embeddings')
    if embeddings is not None:
      np.save(str(out_path.with_suffix('.embed.npy')), embeddings)

    json_str = json.dumps(result, indent=2)
    json_str = compact_json_number_array(json_str)
    out_path.with_suffix('.json').write_text(json_str)



import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import matplotlib.gridspec as gridspec
import librosa.feature
import demucs.separate

from typing import Union, List, Mapping

from .typings import AnalysisResult, PathLike
from .utils import mkpath

HARMONIX_COLORS = {
  'start': 'black',
  'end': 'black',
  'intro': 1,
  'outro': 1,
  'break': 2,
  'bridge': 2,
  'inst': 3,
  'solo': 3,
  'verse': 4,
  'chorus': 5,
}


def visualize(
  results: Union[AnalysisResult, List[AnalysisResult]],
  out_dir: PathLike = None,
) -> Union[plt.Figure, List[plt.Figure]]:
  return_list = True
  if not isinstance(results, list):
    return_list = False
    results = [results]

  figs = [_plot(r) for r in results]

  # Save figures to out_dir
  if out_dir is not None:
    out_dir = mkpath(out_dir)
    out_dir.mkdir(parents=True, exist_ok=True)
    for result, fig in zip(results, figs):
      fig.savefig(out_dir / f'{result.path.stem}.pdf', bbox_inches='tight')

  if not return_list:
    return figs[0]
  return figs


def _plot(
  result: AnalysisResult,
  colors: Mapping[str, int] = None,
  color_map: str = 'viridis',
):
  if colors is None:
    colors = HARMONIX_COLORS

  sr = 44100
  y = demucs.separate.load_track(result.path, 1, sr).numpy()
  # y, sr = librosa.load(result.path, sr=None, mono=True)
  rms = librosa.feature.rms(y=y, frame_length=4096, hop_length=1024)[0]

  fig = plt.figure(figsize=(12, 2))
  gs = gridspec.GridSpec(2, 1, height_ratios=[2, 1])
  ax0 = plt.subplot(gs[0])
  ax1 = plt.subplot(gs[1])

  ax0.plot(rms, color='black', linewidth=1)
  ax0.set_xlim(0, len(rms) - 1)
  ax0.set_ylim(0, None)
  ax0.set_ylabel('RMS')
  ax0.set_xticks([])

  cmap = plt.get_cmap(color_map)
  max_color = max(c for c in colors.values() if not isinstance(c, str))
  min_color = min(c for c in colors.values() if not isinstance(c, str))
  for segment in result.segments:
    color = colors[segment.label]
    if not isinstance(color, str):
      color = (color - min_color) / (max_color - min_color)
      color = cmap(color)

    ax1.axvspan(segment.start, segment.end, color=color)
    ax1.axvline(segment.start, color='black', linewidth=1)
    if segment.label not in ['start', 'end']:
      ax1.text(
        (segment.end - segment.start) / 2 + segment.start,
        0.5,
        segment.label,
        color=_get_text_color(color),
        fontsize=12,
        weight='bold',
        horizontalalignment='center',
        verticalalignment='center',
      )
  ax1.set_xlim(0, result.segments[-1].end)
  tick_segments = [s for s in result.segments if s.label != 'start']
  ax1.set_xticks(
    [s.start for s in tick_segments],
    [f'{round(s.start // 60)}:{round(s.start % 60):02}' for s in tick_segments],
  )
  ax1.set_xlabel('Time (min:sec)')
  ax1.set_yticks([])

  # set title
  ax0.set_title(result.path.name)
  fig.tight_layout()
  fig.subplots_adjust(hspace=0)

  return fig


def _get_text_color(bg_color):
  bg_color = mcolors.to_rgb(bg_color)
  luminance = 0.299 * bg_color[0] + 0.587 * bg_color[1] + 0.114 * bg_color[2]

  if luminance > 0.5:
    return 'black'
  else:
    return 'white'



import numpy as np
import torch
import json

from os import PathLike
from pathlib import Path
from typing import List, Dict, Optional, Union
from dataclasses import dataclass
from numpy.typing import NDArray
from .utils import mkpath

PathLike = Union[str, PathLike]


@dataclass
class AllInOneOutput:
  logits_beat: torch.FloatTensor = None
  logits_downbeat: torch.FloatTensor = None
  logits_section: torch.FloatTensor = None
  logits_function: torch.FloatTensor = None
  embeddings: torch.FloatTensor = None


@dataclass
class Segment:
  start: float
  end: float
  label: str


@dataclass
class AnalysisResult:
  path: Path
  bpm: int
  beats: List[float]
  downbeats: List[float]
  beat_positions: List[int]
  segments: List[Segment]
  activations: Optional[Dict[str, NDArray]] = None
  embeddings: Optional[NDArray] = None

  @staticmethod
  def from_json(path: PathLike):
    path = mkpath(path)
    with open(path, 'r') as f:
      data = json.load(f)

    result = AnalysisResult(
      path=mkpath(data['path']),
      bpm=data['bpm'],
      beats=data['beats'],
      downbeats=data['downbeats'],
      beat_positions=data['beat_positions'],
      segments=[Segment(**seg) for seg in data['segments']],
    )

    activ_path = path.with_suffix('.activ.npz')
    embed_path = path.with_suffix('.embed.npy')
    if activ_path.is_file():
      activs = np.load(activ_path)
      result.activations = {key: activs[key] for key in activs.files}
    if embed_path.is_file():
      result.embeddings = np.load(embed_path)

    return result



import torch

from madmom.features.downbeats import DBNDownBeatTrackingProcessor
from ..typings import AllInOneOutput
from ..config import Config


def postprocess_metrical_structure(
  logits: AllInOneOutput,
  cfg: Config,
):
  postprocessor_downbeat = DBNDownBeatTrackingProcessor(
    beats_per_bar=[3, 4],
    threshold=cfg.best_threshold_downbeat,
    fps=cfg.fps,
  )

  raw_prob_beats = torch.sigmoid(logits.logits_beat[0])
  raw_prob_downbeats = torch.sigmoid(logits.logits_downbeat[0])

  # Transform the raw probabilities into activations indicating:
  # 1. beat but not downbeat
  # 2. downbeat
  # 3. nothing
  # for madmom's DBNDownBeatTrackingProcessor
  activations_beat = raw_prob_beats
  activations_downbeat = raw_prob_downbeats
  activations_no_beat = 1. - activations_beat
  activations_no_downbeat = 1. - activations_downbeat
  activations_no = (activations_no_beat + activations_no_downbeat) / 2.
  activations_xbeat = torch.maximum(torch.tensor(1e-8), activations_beat - activations_downbeat)
  activations_combined = torch.stack([activations_xbeat, activations_downbeat, activations_no], dim=-1)
  activations_combined /= activations_combined.sum(dim=-1, keepdim=True)
  activations_combined = activations_combined.cpu().numpy()

  pred_downbeat_times = postprocessor_downbeat(activations_combined[:, :2])

  beats = pred_downbeat_times[:, 0]
  beat_positions = pred_downbeat_times[:, 1]
  downbeats = pred_downbeat_times[beat_positions == 1., 0]

  beats = beats.tolist()
  downbeats = downbeats.tolist()
  beat_positions = beat_positions.astype('int').tolist()

  return {
    'beats': beats,
    'downbeats': downbeats,
    'beat_positions': beat_positions,
  }



import numpy as np
import librosa
import demucs.separate

from typing import Union, List, Tuple

import torch
from numpy.typing import NDArray
from .typings import AnalysisResult, PathLike, Segment
from .utils import mkpath


def sonify(
  results: Union[AnalysisResult, List[AnalysisResult]],
  out_dir: PathLike = None,
) -> Union[Tuple[NDArray, float], List[Tuple[NDArray, float]]]:
  return_list = True
  if not isinstance(results, list):
    return_list = False
    results = [results]

  sonifs = [_sonify(r) for r in results]

  if out_dir is not None:
    out_dir = mkpath(out_dir)
    out_dir.mkdir(exist_ok=True, parents=True)
    for r, (y, sr) in zip(results, sonifs):
      demucs.separate.save_audio(
        wav=torch.from_numpy(y),
        path=out_dir / f'{r.path.stem}.sonif{r.path.suffix}',
        samplerate=sr,
      )
      # sf.write(
      #   file=out_dir / f'{r.path.stem}.wav',
      #   data=y.T,
      #   samplerate=sr,
      # )

  if not return_list:
    return sonifs[0]
  return sonifs


def _sonify(result: AnalysisResult) -> Tuple[NDArray, float]:
  sr = 44100
  y = demucs.separate.load_track(result.path, 2, sr).numpy()
  # y, sr = librosa.load(result.path, sr=None, mono=False)

  length = y.shape[-1]
  metronome = _sonify_metronome(result, length, sr)
  boundaries = _sonify_boundaries(result.segments, length, sr)

  mixed = y + metronome + boundaries
  mixed = np.clip(mixed, -1, 1)

  return mixed, sr


def _sonify_metronome(
  result: AnalysisResult,
  length: int,
  sr: float = 44100,
):
  # Exclude downbeats from beats.
  downbeats = np.asarray(result.downbeats)
  beats = np.asarray(result.beats)
  dists = np.abs(downbeats[:, np.newaxis] - beats).min(axis=0)
  beats = beats[dists > 0.03]

  clicks_beat = librosa.clicks(
    times=beats,
    sr=sr,
    click_freq=1500,
    click_duration=0.1,
    length=length,
  )
  clicks_downbeat = librosa.clicks(
    times=downbeats,
    sr=sr,
    click_freq=3000,
    click_duration=0.1,
    length=length,
  )

  return clicks_beat + clicks_downbeat


def _sonify_boundaries(
  segments: List[Segment],
  length: int,
  sr: float = 44100,
  riser_freq_start: float = 40.0,
  riser_freq_end: float = 4000.0,
  riser_duration: float = 4.0,
  num_clicks: int = 25,
):
  click_offsets = -np.geomspace(riser_duration, 0.1, num_clicks, endpoint=False)
  boundaries = [segment.start for segment in segments if segment.label not in ['start', 'end']]
  click_times = click_offsets[:, np.newaxis] + boundaries
  click_times = click_times.T.flatten()
  click_freqs = np.geomspace(riser_freq_start, riser_freq_end, num_clicks)
  click_freqs = np.tile(click_freqs, len(boundaries))

  click_freqs = click_freqs[click_times > 0]
  click_times = click_times[click_times > 0]

  y = np.zeros((length,), dtype='float32')
  for t, f in zip(click_times, click_freqs):
    click = _synthesize_click(sr, f)
    y[int(t * sr):int(t * sr) + len(click)] += click

  for segment in segments:
    if segment.label in ['start', 'end']:
      continue
    drop = _synthesize_drop(sr)
    drop_time = segment.start
    y[int(drop_time * sr):int(drop_time * sr) + len(drop)] += drop

  return y


def _synthesize_click(
  sr: float = 44100,
  click_freq: float = 1000.0,
  click_duration: float = 0.1,
):
  angular_freq = 2 * np.pi * click_freq / float(sr)
  click = np.logspace(0, -10, num=int(sr * click_duration), base=2.0)
  click *= np.sin(angular_freq * np.arange(len(click)))
  return click


def _synthesize_drop(
  sr: float = 44100,
  drop_freq_start: float = 4000.0,
  drop_freq_end: float = 40.0,
  drop_duration: float = 0.5,
):
  freqs = np.geomspace(drop_freq_start, drop_freq_end, int(sr * drop_duration))
  drop = np.sin(2 * np.pi * freqs.cumsum() / sr)
  drop *= np.logspace(0, -4, num=int(sr * drop_duration), base=2.0)
  return drop



import numpy as np
import torch.nn.functional as F
import torch
import librosa
from typing import Union
from scipy.signal import argrelextrema
from scipy.interpolate import interp1d
from numpy.lib.stride_tricks import sliding_window_view
from numpy.typing import NDArray
from ..config import Config

def event_frames_to_time(
  tensor: Union[NDArray, torch.Tensor],
  cfg: Config = None,
  sample_rate: int = None,
  hop_size: int = None,
):
  """
  Args:
    tensor: a binary event tensor with shape (batch, frame)
  """
  assert len(tensor.shape) in (1, 2), 'Input tensor should have 1 or 2 dimensions'
  
  if cfg is not None:
    sample_rate = cfg.sample_rate
    hop_size = cfg.hop_size
  
  if torch.is_tensor(tensor):
    tensor = tensor.cpu().numpy()
  
  original_shape = tensor.shape
  if len(original_shape) == 1:
    tensor = tensor[None, :]
  
  batch_size = tensor.shape[0]
  i_examples, i_frames = np.where(tensor)
  times = librosa.frames_to_time(i_frames, sr=sample_rate, hop_length=hop_size)
  times = [times[i_examples == i] for i in range(batch_size)]
  
  if len(original_shape) == 1:
    times = times[0]
  return times


def local_maxima(tensor, filter_size=41):
  assert len(tensor.shape) in (1, 2), 'Input tensor should have 1 or 2 dimensions'
  assert filter_size % 2 == 1, 'Filter size should be an odd number'
  
  original_shape = tensor.shape
  if len(original_shape) == 1:
    tensor = tensor.unsqueeze(0)
  
  # Pad the input array with the minimum value
  padding = filter_size // 2
  padded_arr = F.pad(tensor, (padding, padding), mode='constant', value=-torch.inf)
  
  # Create a rolling window view of the padded array
  rolling_view = padded_arr.unfold(1, filter_size, 1)
  
  # Find the indices of the local maxima
  center = filter_size // 2
  local_maxima_mask = torch.eq(rolling_view[:, :, center], torch.max(rolling_view, dim=-1).values)
  local_maxima_indices = local_maxima_mask.nonzero()
  
  # Initialize a new PyTorch tensor with zeros and the same shape as the input tensor
  output_arr = torch.zeros_like(tensor)
  
  # Set the local maxima values in the output tensor
  output_arr[local_maxima_mask] = tensor[local_maxima_mask]
  
  output_arr = output_arr.reshape(original_shape)
  
  return output_arr, local_maxima_indices


def median_filter_1d(tensor, filter_size=401):
  assert len(tensor.shape) in (1, 2), 'Input tensor should have 1 or 2 dimensions'
  assert filter_size % 2 == 1, 'Filter size should be an odd number'
  
  original_shape = tensor.shape
  original_dtype = tensor.dtype
  if len(original_shape) == 1:
    tensor = tensor.unsqueeze(0)
  
  # Pad the input array with the minimum value
  padding = filter_size // 2
  padded_arr = F.pad(tensor.float(), (padding, padding), mode='reflect')
  padded_arr = padded_arr.to(original_dtype)
  
  # Create a rolling window view of the padded array
  rolling_view = padded_arr.unfold(1, filter_size, 1)
  
  # Compute the median along the new dimension
  output_arr, _ = torch.median(rolling_view, dim=-1)
  
  output_arr = output_arr.reshape(original_shape)
  
  return output_arr


def local_maxima_numpy(arr, order=20):
  is_batch = len(arr.shape) == 2
  if is_batch:
    return np.stack([local_maxima_numpy(x, order) for x in arr])
  
  # Define a comparison function for argrelextrema to find local maxima
  compare_func = np.greater
  
  # Find the indices of the local maxima
  local_maxima_indices = argrelextrema(arr, compare_func, order=order)
  
  # Initialize a new numpy array with zeros and the same shape as the input array
  output_arr = np.zeros_like(arr)
  
  # Set the local maxima values in the output array
  output_arr[local_maxima_indices] = arr[local_maxima_indices]
  
  return output_arr


def binary_to_sawtooth(input_tensor: torch.FloatTensor):
  assert len(input_tensor.shape) in [1, 2], 'Input tensor must be 1D or 2D (batched 1D)'
  if len(input_tensor.shape) == 2:
    return torch.stack([binary_to_sawtooth(x) for x in input_tensor])
  
  device = input_tensor.device
  ones_indices = torch.nonzero(input_tensor).flatten()
  mean_interval = torch.diff(ones_indices).float().mean().round().int()
  pad_left = max(0, mean_interval - ones_indices[0] - 1)
  pad_right = max(0, mean_interval - (input_tensor.shape[0] - ones_indices[-1]))
  
  segment_lengths = torch.diff(
    torch.cat([
      torch.tensor([-pad_left], device=device),
      ones_indices,
      torch.tensor([input_tensor.shape[0] + pad_right], device=device),
    ])
  )
  sawtooth = [torch.linspace(0, 1, length, device=device) for length in segment_lengths]
  sawtooth = torch.cat(sawtooth)
  sawtooth = sawtooth[pad_left:]
  if pad_right > 0:
    sawtooth = sawtooth[:-pad_right]
  assert input_tensor.shape == sawtooth.shape
  return sawtooth


def quad_interp(input_tensor: torch.FloatTensor):
  assert len(input_tensor.shape) in [1, 2], 'Input tensor must be 1D or 2D (batched 1D)'
  if len(input_tensor.shape) == 2:
    return np.stack([quad_interp(x) for x in input_tensor])
  
  input_arr = input_tensor.cpu().numpy()
  arange = np.arange(len(input_arr))
  f = interp1d(arange, input_arr, kind='quadratic')
  output_arr = f(arange)
  return output_arr


def estimate_tempo_from_beats(pred_beat_times):
  beat_interval = np.diff(pred_beat_times)
  bpm = 60. / beat_interval
  # TODO: quadratic interpolation
  bpm = bpm.round()
  bincount = np.bincount(bpm.astype(int))
  bpm_range = np.arange(len(bincount))
  bpm_strength = bincount / bincount.sum()
  bpm_est = np.stack([bpm_range, bpm_strength], axis=-1)
  bpm_est = bpm_est[np.argsort(bpm_strength)[::-1]]
  bpm_est = bpm_est[bpm_est[:, 1] > 0]
  return bpm_est


def peak_picking(boundary_activation, window_past=12, window_future=6):
  # Find local maxima using a sliding window
  window_size = window_past + window_future
  assert window_size % 2 == 0, 'window_past + window_future must be even'
  window_size += 1

  # Pad boundary_activation
  boundary_activation_padded = np.pad(boundary_activation, (window_past, window_future), mode='constant')
  max_filter = sliding_window_view(boundary_activation_padded, window_size)
  local_maxima = (boundary_activation == np.max(max_filter, axis=-1)) & (boundary_activation > 0)

  # Compute strength values by subtracting the mean of the past and future windows
  past_window_filter = sliding_window_view(boundary_activation_padded[:-(window_future + 1)], window_past)
  future_window_filter = sliding_window_view(boundary_activation_padded[window_past + 1:], window_future)
  past_mean = np.mean(past_window_filter, axis=-1)
  future_mean = np.mean(future_window_filter, axis=-1)
  strength_values = boundary_activation - ((past_mean + future_mean) / 2)

  # Get boundary candidates and their corresponding strength values
  boundary_candidates = np.flatnonzero(local_maxima)
  strength_values = strength_values[boundary_candidates]

  strength_activations = np.zeros_like(boundary_activation)
  strength_activations[boundary_candidates] = strength_values

  return strength_activations
